# -*- coding: utf-8 -*-
"""
Created on Tue Apr  5 15:07:00 2016

@author:
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""

import numpy as np
import matplotlib
import matplotlib.pyplot as plt
#from mpl_toolkits.axes_grid1 import AxesGrid



def shiftedColorMap(cmap, start=0, midpoint=0.5, stop=1.0, name='shiftedcmap'):
    '''
    Function to offset the "center" of a colormap. Useful for
    data with a negative min and positive max and you want the
    middle of the colormap's dynamic range to be at zero

    Input
    -----
      cmap : The matplotlib colormap to be altered
      start : Offset from lowest point in the colormap's range.
          Defaults to 0.0 (no lower ofset). Should be between
          0.0 and `midpoint`.
      midpoint : The new center of the colormap. Defaults to 
          0.5 (no shift). Should be between 0.0 and 1.0. In
          general, this should be  1 - vmax/(vmax + abs(vmin))
          For example if your data range from -15.0 to +5.0 and
          you want the center of the colormap at 0.0, `midpoint`
          should be set to  1 - 5/(5 + 15)) or 0.75
      stop : Offset from highets point in the colormap's range.
          Defaults to 1.0 (no upper ofset). Should be between
          `midpoint` and 1.0.
    '''
    cdict = {
        'red': [],
        'green': [],
        'blue': [],
        'alpha': []
    }

    # regular index to compute the colors
    reg_index = np.linspace(start, stop, 257)

    # shifted index to match the data
    shift_index = np.hstack([
        np.linspace(0.0, midpoint, 128, endpoint=False), 
        np.linspace(midpoint, 1.0, 129, endpoint=True)
    ])

    for ri, si in zip(reg_index, shift_index):
        r, g, b, a = cmap(ri)

        cdict['red'].append((si, r, r))
        cdict['green'].append((si, g, g))
        cdict['blue'].append((si, b, b))
        cdict['alpha'].append((si, a, a))

    newcmap = matplotlib.colors.LinearSegmentedColormap(name, cdict)
    plt.register_cmap(cmap=newcmap)

    return newcmap
    
    
    
    
def plot_contour(ax,x1,y1):
    H, xedges, yedges = np.histogram2d(x1,y1)
    extent = [xedges[0],xedges[-1],yedges[0],yedges[-1]]
    ax.contour(H.T,extent=extent,zorder=10)
    return ax


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  5 15:07:00 2016

@author:
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""

import numpy as np
#from scipy import stats
import time
import os, glob
#from running_median import RunningMedian
#from mpl_toolkits.axes_grid1 import AxesGrid
try:
    import pandas as pd
except ImportError:
    'Warning: module "pandas" not installed.'
        
        
def medsig(a):
    '''Compute median and MAD-estimated scatter of array a'''
#    try:
    med = np.nanmedian(a)
    sig = 1.48 * np.nanmedian(abs(a-med))
#    except:
#        AttributeError
#    else:
#        med = stats.nanmedian(a)
#        sig = 1.48 * stats.nanmedian(abs(a-med))
    return med, sig   
    
  
  
def running_mean(x, N):
    x[np.isnan(x)] = 0. #reset NAN to 0 to calculate the cumulative sum; mimics the 'pandas' behavior
    cumsum = np.cumsum(np.insert(x, 0., 0.)) 
    return 1.*(cumsum[N:] - cumsum[:-N]) / N 
    
    
# 'running_median' DOES NOT AGREE WITH THE PANDAS IMPLEMENTATION 'running_median_pandas'
#def running_median(x, N):
#    return np.array(list(RunningMedian(N, x)))
    
    
def running_mean_pandas(x, N):
    ts = pd.Series(x).rolling(window=N, center=False).mean()
    return ts[~np.isnan(ts)].as_matrix()



def running_median_pandas(x, N):
    ts = pd.Series(x).rolling(window=N, center=False).median()
    return ts[~np.isnan(ts)].as_matrix()


    
def mask_ranges(x, x_min, x_max):
    """"
    Crop out values and indices out of an array x for multiple given ranges x_min to x_max.
    
    Input:
    x: array, 
    x_min: lower limits of the ranges
    x_max: upper limits of the ranges
    
    Output:
    
    
    Example:
    x = np.arange(200)    
    x_min = [5, 25, 90]
    x_max = [10, 35, 110]
    """

    mask = np.zeros(len(x), dtype=bool)
    for i in range(len(x_min)): 
        mask = mask | ((x >= x_min[i]) & (x <= x_max[i]))
    ind_mask = np.arange(len(mask))[mask]
    
    return x[mask], ind_mask, mask 
    
    

def mystr(x,digits=0):
    if np.isnan(x): return '.'
    elif digits==0: return str(int(round(x,digits)))
    else: return str(round(x,digits))
    
    

def version_control(files='*.py', printing=True):
    #    last_created_file = max(glob.iglob(files), key=os.path.getctime)
    last_updated_file = max(glob.iglob(files), key=os.path.getmtime)
    if printing == True: 
    #    print "# Last created script: %s, %s" % ( last_created_file, time.ctime(os.path.getmtime(last_created_file)) )
    #    print "# Last updated script: %s, %s" % ( last_updated_file, time.ctime(os.path.getmtime(last_updated_file)) )
        print "# Last update: %s" % time.ctime(os.path.getmtime(last_updated_file))
        

def merge_two_dicts(x, y):
    '''Given two dicts, merge them into a new dict as a shallow copy.'''
    z = x.copy()
    z.update(y)
    return z
        
        
def table_view(dic):
    from astropy.table import Table 
    dic_table = {}
    subkeys = ['OBJ_ID', 'SYSREM_FLUX3_median', 'PERIOD', 'DEPTH', 'WIDTH', 'NUM_TRANSITS']
    for key in subkeys:
            dic_table[key] = dic[key]
    dic_table = Table(dic_table)
    dic_table = dic_table[subkeys]
    print dic_table


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 11 18:28:27 2016

@author:
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""

# http://www.bdnyc.org/2012/10/decimal-deg-to-hms/


def deg2HMS(ra='', dec='', round=False):
  RA, DEC, rs, ds = '', '', '', ''
  if dec:
    if str(dec)[0] == '-':
      ds, dec = '-', abs(dec)
    deg = int(dec)
    decM = abs(int((dec-deg)*60))
    if round:
      decS = int((abs((dec-deg)*60)-decM)*60)
    else:
      decS = (abs((dec-deg)*60)-decM)*60
    DEC = '{0}{1} {2} {3}'.format(ds, deg, decM, decS)
  
  if ra:
    if str(ra)[0] == '-':
      rs, ra = '-', abs(ra)
    raH = int(ra/15)
    raM = int(((ra/15)-raH)*60)
    if round:
      raS = int(((((ra/15)-raH)*60)-raM)*60)
    else:
      raS = ((((ra/15)-raH)*60)-raM)*60
    RA = '{0}{1} {2} {3}'.format(rs, raH, raM, raS)
  
  if ra and dec:
    return (RA, DEC)
  else:
    return RA or DEC


# -*- coding: utf-8 -*-
"""
Created on Fri Aug  5 17:55:21 2016

@author:
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""

import numpy as np
#import matplotlib.pyplot as plt
from My_Utils import medsig


def rebin_err(t, f, ferr=None, dt = 0.02, ferr_type='medsig', ferr_style='std'):
    """
    @written by Ed Gillen
    The standard rebin function but also dealing with errors
    on the individual data points being binned.
    ferr_type:
        'medsig'
        'meanstd'
    ferr_style:
        'std'
        'sem' = std / sqrt(N)
    """
    treg = np.r_[t.min():t.max():dt]
    nreg = len(treg)
    freg = np.zeros(nreg) + np.nan
#    if ferr!=None:
    freg_err = np.ma.zeros(nreg) + np.nan
    for i in np.arange(nreg):
        l = (t >= treg[i]) * (t < treg[i] + dt)
        if l.any():
            treg[i] = np.ma.mean(t[l])
            if ferr==None:
                if ferr_type == 'medsig':
                    freg[i], freg_err[i] = medsig(f[l])
                else:
                    freg[i] = np.nanmean(f[l])
                    freg_err[i] = np.nanstd(f[l])
                    
                if ferr_style == 'sem':
                    freg_err[i] /= np.sqrt( len(f[l]) )
            else:
                freg[i], freg_err[i] = weighted_avg_and_std( f[l], np.ma.array([1/float(x) for x in ferr[l]]) )
    l = np.isfinite(freg)
#    if ferr==None:
#        return treg[l], freg[l]
    return treg[l], freg[l], freg_err[l]
    
    
    
def weighted_avg_and_std(values, weights):
    """
    Return the weighted average and standard deviation.
    values, weights -- Numpy ndarrays with the same shape.
    """
    average = np.ma.average(values, weights=weights)
    variance = np.ma.average((values-average)**2, weights=weights)  # Fast and numerically precise
    return [average, np.sqrt(variance)]

    
    
def calc_phase(hjd, P, Tprim):
    return ((hjd - Tprim) % P) / P
    
    
    
def phase_fold(time, flux, P, Tprim, bin_width_phase=0.02, ferr_type='medsig', ferr_style='std'):
    phi = calc_phase( time, P, Tprim )
    phi[ phi>0.75 ] -= 1.
    phase, phaseflux, phaseflux_err = rebin_err( phi, flux, None, dt=bin_width_phase, ferr_type=ferr_type, ferr_style=ferr_style )
    return phase, phaseflux, phaseflux_err, phi


import numpy as np
from My_Utils import running_mean, running_median, running_mean_pandas, running_median_pandas
from binning import binning1D


np.random.seed(1)
data = np.random.random(20) * np.random.randint(1,20,20)
x = 5.

mean,_ = binning1D(data, x, setting='mean', normalize=False) 
median,_ = binning1D(data, x, setting='median', normalize=False) 
runmean = running_mean(data, x)
runmedian = running_median(data, x)
runmean_pandas = running_mean_pandas(data, x)
runmedian_pandas = running_median_pandas(data, x)




print data
print '---'
print 'mean', mean, '>>>', np.std(mean)
print 'runmean', runmean, '>>>', np.std(runmean)
print 'runmean_pandas', runmean_pandas, '>>>', np.std(runmean_pandas)
print '---'
print 'median', median, '>>>', np.std(median)
print 'runmedian', runmedian, '>>>', np.std(runmean)
print 'runmedian_pandas', runmedian_pandas, '>>>', np.std(runmedian_pandas)

# RUNMEDIAN DOES NOT AGREE WITH THE PANDAS IMPLEMENTATION


##::: COMPARISON 1 & 2:
##::: (copy code into ipython console)
#x = np.random.random(100000)
#N = 1000
#%timeit np.convolve(x, np.ones((N,))/N, mode='valid')
##10 loops, best of 3: 172 ms per loop
#
#%timeit running_mean(x, N)
##100 loops, best of 3: 6.72 ms per loop
#
#%timeit pd.rolling_mean(x, N)[N-1:]
##100 loops, best of 3: 4.74 ms per loop
#
#np.allclose(pd.rolling_mean(x, N)[N-1:], running_mean(x, N))
##True



# -*- coding: utf-8 -*-
"""
Created on Wed Jun 22 20:36:09 2016

@author:
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""

import numpy as np
import matplotlib.pyplot as plt


#::: if only 1 object is contained in dic
def set_nan(dic):
    ###### REMOVE BROKEN ITEMS #######
    #::: nan
    ind_broken = np.where( dic['SYSREM_FLUX3'] == 0. )
    if 'SYSREM_FLUX3' in dic: dic['SYSREM_FLUX3'][ind_broken] = np.nan
#    dic['HJD'][ind_broken] = np.nan #this is not allowed to be set to nan!!! Otherwise the binning will be messed up!!!
    if 'CCDX' in dic: dic['CCDX'][ind_broken] = np.nan
    if 'CCDY' in dic: dic['CCDY'][ind_broken] = np.nan
    if 'CENTDX' in dic: dic['CENTDX'][ind_broken] = np.nan
    if 'CENTDY' in dic: dic['CENTDY'][ind_broken] = np.nan
    return dic
    

#::: if multiple objects are contained in dic
def set_nan_multi(dic):
    ###### REMOVE BROKEN ITEMS #######
    #::: nan
    N_obj = dic['SYSREM_FLUX3'].shape[0]
    for obj_nr in range(N_obj):
        ind_broken = np.where( dic['SYSREM_FLUX3'][obj_nr] == 0. )
        if 'SYSREM_FLUX3' in dic: dic['SYSREM_FLUX3'][obj_nr,ind_broken] = np.nan
    #    dic['HJD'][ind_broken] = np.nan #this is not allowed to be set to nan!!! Otherwise the binning will be messed up!!!
        if 'CCDX' in dic: dic['CCDX'][obj_nr,ind_broken] = np.nan
        if 'CCDY' in dic: dic['CCDY'][obj_nr,ind_broken] = np.nan
        if 'CENTDX' in dic: dic['CENTDX'][obj_nr,ind_broken] = np.nan
        if 'CENTDY' in dic: dic['CENTDY'][obj_nr,ind_broken] = np.nan
    return dic


"""
http://code.activestate.com/recipes/576930/
"""



from random import random
from math import log
import numpy as np




class Node(object):
    __slots__ = 'value', 'next', 'width'
    def __init__(self, value, next, width):
        self.value, self.next, self.width = value, next, width

class End(object):
    'Sentinel object that always compares greater than another object'
    def __cmp__(self, other):
        return 1

NIL = Node(End(), [], [])               # Singleton terminator node

class IndexableSkiplist:
    'Sorted collection supporting O(lg n) insertion, removal, and lookup by rank.'

    def __init__(self, expected_size=100):
        self.size = 0
        self.maxlevels = int(1 + log(expected_size, 2))
        self.head = Node('HEAD', [NIL]*self.maxlevels, [1]*self.maxlevels)

    def __len__(self):
        return self.size

    def __getitem__(self, i):
        node = self.head
        i += 1
        for level in reversed(range(self.maxlevels)):
            while node.width[level] <= i:
                i -= node.width[level]
                node = node.next[level]
        return node.value

    def insert(self, value):
        # find first node on each level where node.next[levels].value > value
        chain = [None] * self.maxlevels
        steps_at_level = [0] * self.maxlevels
        node = self.head
        for level in reversed(range(self.maxlevels)):
            while node.next[level].value <= value:
                steps_at_level[level] += node.width[level]
                node = node.next[level]
            chain[level] = node

        # insert a link to the newnode at each level
        d = min(self.maxlevels, 1 - int(log(random(), 2.0)))
        newnode = Node(value, [None]*d, [None]*d)
        steps = 0
        for level in range(d):
            prevnode = chain[level]
            newnode.next[level] = prevnode.next[level]
            prevnode.next[level] = newnode
            newnode.width[level] = prevnode.width[level] - steps
            prevnode.width[level] = steps + 1
            steps += steps_at_level[level]
        for level in range(d, self.maxlevels):
            chain[level].width[level] += 1
        self.size += 1

    def remove(self, value):
        # find first node on each level where node.next[levels].value >= value
        chain = [None] * self.maxlevels
        node = self.head
        for level in reversed(range(self.maxlevels)):
            while node.next[level].value < value:
                node = node.next[level]
            chain[level] = node
        if value != chain[0].next[0].value:
            raise KeyError('Not Found')

        # remove one link at each level
        d = len(chain[0].next[0].next)
        for level in range(d):
            prevnode = chain[level]
            prevnode.width[level] += prevnode.next[level].width[level] - 1
            prevnode.next[level] = prevnode.next[level].next[level]
        for level in range(d, self.maxlevels):
            chain[level].width[level] -= 1
        self.size -= 1

    def __iter__(self):
        'Iterate over values in sorted order'
        node = self.head.next[0]
        while node is not NIL:
            yield node.value
            node = node.next[0]

from collections import deque
from itertools import islice

class RunningMedian:
    'Fast running median with O(lg n) updates where n is the window size'

    def __init__(self, n, iterable):
        self.it = iter(iterable)
        self.queue = deque(islice(self.it, n))
        self.skiplist = IndexableSkiplist(n)
        for elem in self.queue:
            self.skiplist.insert(elem)

    def __iter__(self):
        queue = self.queue
        skiplist = self.skiplist
        midpoint = len(queue) // 2
        yield skiplist[midpoint]
        for newelem in self.it:
            oldelem = queue.popleft()
            skiplist.remove(oldelem)
            queue.append(newelem)
            skiplist.insert(newelem)
            yield skiplist[midpoint]



#def running_median(x, N):
#    return np.array(RunningMedian(N, x))
    
    
    

if __name__ == '__main__':

    #########################################################################
    # Demonstrate the RunningMedian() class
    # Compare results to alternative class implemented using a regular list

    from bisect import insort
    from random import randrange

    class RunningMedianSlow:
        'Slow running-median with O(n) updates where n is the window size'

        def __init__(self, n, iterable):
            self.it = iter(iterable)
            self.queue = deque(islice(self.it, n))
            self.sortedlist = sorted(self.queue)

        def __iter__(self):
            queue = self.queue
            sortedlist = self.sortedlist
            midpoint = len(queue) // 2
            yield sortedlist[midpoint]
            for newelem in self.it:
                oldelem = queue.popleft()
                sortedlist.remove(oldelem)
                queue.append(newelem)
                insort(sortedlist, newelem)
                yield sortedlist[midpoint]

    M, N, window = 5000, 8000, 1001
    data = [randrange(M) for i in range(N)]
    result = list(RunningMedian(window, data))
    expected = list(RunningMedianSlow(window, data))
    assert result == expected
    print 'Successful test of RunningMedian() with', N,
    print 'items and a window of size', window, '\n'
    print result
    print type(RunningMedian(window, data))


    #########################################################################
    # Demonstrate and test the IndexableSkiplist() collection class

    from random import shuffle

    demo = list('FOURSQUARE')
    excluded = list('XZ')

    print 'Building an indexable skiplist containing %r:' % ''.join(demo)
    s = IndexableSkiplist(len(demo))                        # exercise __init__()
    shuffle(demo)
    for char in demo:
        s.insert(char)                                      # check insert()
        print 'Adding: ', char, list(s), len(s)

    for char in excluded:
        print 'Attempting to remove:', char, '-->',
        try:
            s.remove(char)                                  # check remove() item not in skiplist
        except KeyError:
            print 'Successfully not found'

    assert len(s) == len(demo)                              # check __len__()
    assert list(s) == sorted(demo)                          # check __iter__()
    assert [s[i] for i in range(len(s))] == sorted(demo)    # check __getitem__()

    shuffle(demo)
    for char in demo:
        s.remove(char)                                      # check remove() item in skiplist
        print 'Removing: ', char, list(s), len(s)
    assert list(s) == [] and len(s) == 0


    #############################################################################
    # Show the IndexableSkiplist's internal structure and describe its invariants

    def iter_levels(skiplist, level):
        'Iterate through nodes on a given level.'
        node = skiplist.head
        while node is not NIL:
            yield node
            node = node.next[level]

    demo = list('HOPSCOTCH')                    # create a sample skiplist
    s = IndexableSkiplist(len(demo))
    for char in demo:
        s.insert(char)

    print '\nStructure of an indexable skiplist containing %r:' % ''.join(demo)
    for i in reversed(range(s.maxlevels)):      # show the link structure on each level
        print 'Level %d:  ' % i,
        line = ''
        for n in iter_levels(s, i):
            line += '%s%d %s' % (n.value, n.width[i], '   ' * (n.width[i] - 1))
        print line + 'NIL'

    print '\nNote the following structure invariants:'
    print '* Level 0 has widths all equal to one.'
    print '* The sum of widths on each level == len(skiplist)+1.'
    print '* Entries on higher levels are subsets of the lower levels.'
    print '* Widths of entries on higer levels sum to the widths below.'
    print '    For example, a P --> T link of two steps on one level'
    print '    corresponds to P --> S --> T links of one step each on level 0.'



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 11 15:32:21 2016

@author: 
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""


from setuptools import setup

setup(
    name = 'mytools',      # The name of the PyPI-package.
    packages = ['mytools'],
    version = '1.0.0',    # Update the version number for new releases
    #scripts=['ngtsio'],  # The name of the included script(s), and also the command used for calling it
    description = 'Set of tools for photometric time series data and plotting',
    author = 'Maximilian N. Guenther',
    author_email = 'mg719@cam.ac.uk',
    url = 'https://github.com/MNGuenther/mytools',
    download_url = 'https://github.com/MNGuenther/mytools/releases',
    classifiers = [],
    install_requires=['numpy>=1.10','matplotlib>0.1','scipy>=0.1','pandas>=0.1']
    )






# -*- coding: utf-8 -*-
"""
Created on Sun Mar 13 21:18:27 2016

@author:
Maximilian N. Guenther
Battcock Centre for Experimental Astrophysics,
Cavendish Laboratory,
JJ Thomson Avenue
Cambridge CB3 0HE
Email: mg719@cam.ac.uk
"""

import numpy as np
import matplotlib.pyplot as plt
#from scipy import stats #needed for stats.nanmean/median on ngtshead



######################################################################
# BINNING WITHOUT TIME GAPS 
# !!! DO NOT USE FOR COMBINING DIFFERENT NIGHTS !!!
######################################################################

def binning1D(arr, bin_width, setting='mean', normalize=False):
    """ WARNING: this does not respect boundaries between different night; 
    will average data from different nights"""
    N_time = len(arr)
    N_bins = np.int64(np.ceil(1.*N_time / bin_width))
    binarr, binarr_err = np.zeros((2,N_bins))
    bin_width = int(bin_width)
    if setting=='mean':
        for nn in range(N_bins):
            binarr[nn] = np.nanmean(arr[nn*bin_width:(nn+1)*bin_width])
            binarr_err[nn] = np.nanstd(arr[nn*bin_width:(nn+1)*bin_width])
    if setting=='median':
        for nn in range(N_bins):
            binarr[nn] = np.nanmedian(arr[nn*bin_width:(nn+1)*bin_width])
            binarr_err[nn] = 1.48 * np.nanmedian(abs(arr[nn*bin_width:(nn+1)*bin_width] - binarr[nn]))
    if normalize==True:
        med = np.nanmedian(binarr)
        binarr /= med
        binarr_err /= med
        
    return binarr, binarr_err
    
    
    
def binning2D(arr, bin_width, setting='mean', normalize=False, axis=1):
    #arr being 2D array, with objs on x and time stamps on y
    """ WARNING: this does not respect boundaries between different night; 
    will average data from different nights"""
    N_time = arr.shape[1]
#    print N
    N_objs = arr.shape[0]
#    print N_objs
    N_bins = np.int64(np.ceil(1.*N_time / bin_width))
#    print N_bins
    binarr, binarr_err = np.zeros((2,N_objs,N_bins))
#    print arr.shape
#    print binarr.shape
    bin_width = int(bin_width)
    if setting=='mean':
        for nn in range(N_bins):
            binarr[:,nn] = np.nanmean(arr[:,nn*bin_width:(nn+1)*bin_width], axis=axis)
            binarr_err[:,nn] = np.nanstd(arr[:,nn*bin_width:(nn+1)*bin_width], axis=axis)
    if setting=='median':
        for nn in range(N_bins):
            binarr[:,nn] = np.nanmedian(arr[:,nn*bin_width:(nn+1)*bin_width], axis=axis)
            binarr_err[:,nn] = 1.48 * np.nanmedian(abs(arr[:,nn*bin_width:(nn+1)*bin_width] - binarr[:,nn]))
    if normalize==True:
        med = np.nanmedian(binarr)
        binarr /= med
        binarr_err /= med
        
#    print arr.shape
#    print binarr.shape
        
    return binarr, binarr_err
    
    

######################################################################
# BINNING WITH TIME GAPS 
# !!! USE THIS FOR COMBINING DIFFERENT NIGHTS !!!
######################################################################   
  
def bin_edge_indices(time1D, bin_width, timegap, N_time):
    """ DETERMINE ALL THE BIN-EDGE-INDICES (TO NOT BIN OVER DIFFERENT NIGHTS)"""
    """ this currently relies on the fact that timestamps for all are approximately the same 
    (given for the case of a HJD array that represents MJD values with small corrections)"""

#    ind_start_of_night = np.append( 0 , np.where( np.diff(time) > timegap )[0] + 1 )
    ind_end_of_night = np.append( np.where( np.diff(time1D) > timegap )[0], len(np.diff(time1D)-1 ) )
    N_nights = len(ind_end_of_night)
    
    first_ind = [0]
    last_ind = []
    i = 0    
#    j = 0
    while ((first_ind[-1] < N_time) & (i < N_nights) ):
        if (first_ind[-1]+bin_width) < ind_end_of_night[i]:
            last_ind.append( first_ind[-1] + bin_width )
        else:
            last_ind.append( ind_end_of_night[i] )
            i += 1
        first_ind.append( last_ind[-1] + 1 )
#        j += 1
    
    del first_ind[-1]  
    
    return first_ind, last_ind
    


def binning1D_per_night(time, arr, bin_width, timegap=3600, setting='mean', normalize=False):
    """ If time and arr are 1D arrays """
    
    N_time = len(arr)
    bin_width = int(bin_width)
    
    first_ind, last_ind = bin_edge_indices(time, bin_width, timegap, N_time)
    
    N_bins = len(first_ind)
    bintime, binarr, binarr_err = np.zeros((3,N_bins)) * np.nan
    
    if setting=='mean':
        for nn in range(N_bins):
            #skip no/single data points
            if last_ind[nn] > first_ind[nn]:
                bintime[nn] = np.nanmean( time[first_ind[nn]:last_ind[nn]] )
                #skip All-NAN slices (i.e. where all flux data is masked)
                if ( np.isnan(arr[first_ind[nn]:last_ind[nn]]).all() == False ):
                    binarr[nn] = np.nanmean( arr[first_ind[nn]:last_ind[nn]] )
                    binarr_err[nn] = np.nanstd( arr[first_ind[nn]:last_ind[nn]] )
    elif setting=='median':
        for nn in range(N_bins):
            #skip no/single data points
            if (last_ind[nn] > first_ind[nn]): 
                bintime[nn] = np.nanmedian( time[first_ind[nn]:last_ind[nn]] )
                #skip All-NAN slices (i.e. where all flux data is masked)
                if ( np.isnan(arr[first_ind[nn]:last_ind[nn]]).all() == False ):
                    binarr[nn] = np.nanmedian( arr[first_ind[nn]:last_ind[nn]] )
                    binarr_err[nn] = 1.48 * np.nanmedian( abs(arr[first_ind[nn]:last_ind[nn]] - binarr[nn]) )
    if normalize==True:
        med = np.nanmedian(binarr)
        binarr /= med
        binarr_err /= med
            
    return bintime, binarr, binarr_err     
    


def binning2D_per_night(time, arr, bin_width, timegap=3600, setting='mean', normalize=False, axis=1):
    """ If time and arr are each a 2D array, with different objs on x and different time stamps on y"""
    """ this currently relies on the fact that timestamps for all are approximately the same 
    (given for the case of a HJD array that represents MJD values with small corrections)"""
    N_time = arr.shape[1]
    N_objs = arr.shape[0]
    
    bin_width = int(bin_width)
    
    first_ind, last_ind = bin_edge_indices(time[0,:], bin_width, timegap, N_time)
    
    N_bins = len(first_ind)
    bintime, binarr, binarr_err = np.zeros((3,N_objs,N_bins))
    
    if setting=='mean':
        for nn in range(N_bins):
            bintime[:,nn] = np.nanmean( time[:,first_ind[nn]:last_ind[nn]], axis=axis )
            binarr[:,nn] = np.nanmean( arr[:,first_ind[nn]:last_ind[nn]], axis=axis )
            binarr_err[:,nn] = np.nanstd( arr[:,first_ind[nn]:last_ind[nn]], axis=axis )
    elif setting=='median':
        for nn in range(N_bins):
            bintime[:,nn] = np.nanmedian( time[:,first_ind[nn]:last_ind[nn]], axis=axis )
            binarr[:,nn] = np.nanmedian( arr[:,first_ind[nn]:last_ind[nn]], axis=axis )
            binarr_err[:,nn] = 1.48 * np.nanmedian( abs(arr[:,first_ind[nn]:last_ind[nn]] - binarr[:,nn]) )
    
    if normalize==True:
        med = np.nanmedian(binarr)
        binarr /= med
        binarr_err /= med
            
    return bintime, binarr, binarr_err
               

    
def binning1D_per_night_list(time, arr, bin_width, timegap=3600, setting='mean', normalize=False):
    """ different style of program, same application """
    N = len(time)
    bin_width = int(bin_width)
    
    bintime = []
    binarr = []
    binarr_err = []
    
#    ind_start_of_night = np.append( 0 , np.where( np.diff(time) > timegap )[0] + 1 )
    ind_end_of_night = np.append( np.where( np.diff(time) > timegap )[0], len(np.diff(time)-1 ) )
    N_nights = len(ind_end_of_night)
    first_ind = 0
    i = 0
    
    if setting=='mean':
        while ((first_ind < N) & (i < N_nights) ):
            if (first_ind+bin_width) < ind_end_of_night[i]:
                last_ind = first_ind+bin_width
            else:
                last_ind = ind_end_of_night[i]
                i += 1
                
            bintime.append( np.nanmean( time[first_ind:last_ind] ) )
            binarr.append( np.nanmean( arr[first_ind:last_ind] ) )
            binarr_err.append( np.nanstd(arr[first_ind:last_ind]) )
            first_ind = last_ind + 1
            
    elif setting=='median':       
        while first_ind < N:
            if (first_ind+bin_width) < ind_end_of_night[i]:
                last_ind = first_ind+bin_width
            else:
                last_ind = ind_end_of_night[i]
                i += 1
                
            bintime.append( np.nanmedian( time[first_ind:last_ind] ) )
            binarr.append( np.nanmedian( arr[first_ind:last_ind] ) )
            binarr_err.append( 1.48 * np.nanmedian(abs( arr[first_ind:last_ind] - binarr[-1])) )
            first_ind = last_ind

    bintime = np.array(bintime)
    binarr = np.array(binarr)
    binarr_err = np.array(binarr_err)
    
    if normalize==True:
        med = np.nanmedian(binarr)
        binarr /= med
        binarr_err /= med
        
    return bintime, binarr, binarr_err       
    
    

######################################################################
# MAIN (FOR TESTING)
######################################################################
if __name__ == '__main__':
    
    ######################################################################
    # TEST binning2D_per_night
    ######################################################################
    arr = np.array([[1,2,3,4,5,6,  67,68,64,  -10,-11,-13], \
                   [1,2,3,4,5,6,  24,28,32,  10,11,13]])           
    time = np.array([[1,2,3,4,5,6,   10001,10002,10003,    20001,20002,20003], \
                    [1,2,3,4,5,6.1,   10001,10002.1,10003.3,    20001,20002,20003]])
    
    bintime,binarr, _ = binning2D_per_night(time,arr,6)
    
    plt.figure()
    plt.plot(time,arr,'k.')
    plt.plot(bintime,binarr,'r.')
    
    
    ######################################################################
    # TEST binning1D_per_night
    ######################################################################
    arr = np.array([1,2,3,4,5,6,  67,68,64,  -10,-11,-13])         
    time = np.array([1,2,3,4,5,6,   10001,10002,10003,    20001,20002,20003])
                    
    bintime,binarr, _ = binning1D_per_night(time,arr,6)
    
    plt.figure()
    plt.plot(time,arr,'k.')
    plt.plot(bintime,binarr,'r.')